"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.observe = exports.observeArray = exports.observeWeakMap = exports.observeObj = exports.originalFnObserve = exports.observeFunction = void 0;
var atoms_1 = require("./atoms");
var calculations_1 = require("./calculations");
var mutations_1 = require("./mutations");
var Observable_1 = require("./Observable");
// tslint:disable: no-unsafe-any
function observeFunction(fn, onSet) {
    var createAtomIfNotExists = (0, atoms_1.getAtomCreator)(fn);
    return new Proxy(fn, {
        apply: function (target, thisArg, args) {
            var p = JSON.stringify({ args: args });
            var atom = createAtomIfNotExists(p);
            // function might have side effects, so we don't memorize it
            // TODO: investigate if `mutation` information is sufficient enough
            // to detect when function is pure
            (0, calculations_1.linkToCalculated)(atom);
            return (0, calculations_1.recalculate)(atom, function () { return Reflect.apply(target, thisArg, args); }, observe);
        },
    });
}
exports.observeFunction = observeFunction;
function originalFnObserve(fn, onSet) {
    var createAtomIfNotExists = (0, atoms_1.getAtomCreator)(fn);
    return new Proxy(fn, {
        apply: function (target, thisArg, args) {
            var p = JSON.stringify({ args: args });
            var atom = createAtomIfNotExists(p);
            // function might have side effects, so we don't memorize it
            // TODO: investigate if `mutation` information is sufficient enough
            // to detect when function is pure
            (0, calculations_1.linkToCalculated)(atom);
            return (0, calculations_1.recalculate)(atom, function () { return Reflect.apply(target, thisArg, args); }, observe);
        },
    });
}
exports.originalFnObserve = originalFnObserve;
function observeObj(obj, onSet) {
    var createAtomIfNotExists = (0, atoms_1.getAtomCreator)(obj);
    return new Proxy(obj, {
        // TODO: fix `symbol` type
        get: function (target, p, receiver) {
            // preserving Proxy invariant
            var desc = Reflect.getOwnPropertyDescriptor(target, p);
            if (desc && !desc.configurable && !desc.writable) {
                return Reflect.get(target, p, receiver);
            }
            var atom = createAtomIfNotExists(p);
            return (0, calculations_1.calculate)(atom, function () { return Reflect.get(target, p, receiver); }, observe);
        },
        // TODO: fix `symbol` type
        set: function (target, p, value, receiver) {
            var atom = createAtomIfNotExists(p, value);
            return (0, mutations_1.mutate)(atom, function () { return Reflect.set(target, p, value, receiver); }, onSet);
        },
        // TODO: fix `symbol` type
        defineProperty: function (target, p, attributes) {
            var value = attributes.value;
            createAtomIfNotExists(p, value);
            return Reflect.defineProperty(target, p, attributes);
        },
    });
}
exports.observeObj = observeObj;
function observeWeakMap(obj, onSet) {
    var createAtomIfNotExists = (0, atoms_1.getAtomCreator)(obj);
    return new Proxy(obj, {
        // TODO: fix `symbol` type
        get: function (target, p, receiver) {
            // preserving Proxy invariant
            var desc = Reflect.getOwnPropertyDescriptor(target, p);
            if (desc && !desc.configurable && !desc.writable) {
                return Reflect.get(target, p, receiver);
            }
            var atom = createAtomIfNotExists(p);
            return (p === 'get' || p === 'set' || p === 'has' || p === 'delete')
                ? (0, calculations_1.calculate)(atom, function () { return Reflect.get(obj, p, receiver); }, observe)
                : (0, calculations_1.calculate)(atom, function () { return Reflect.get(target, p, receiver); }, observe);
        },
        // TODO: fix `symbol` type
        set: function (target, p, value, receiver) {
            var atom = createAtomIfNotExists(p, value);
            return (0, mutations_1.mutate)(atom, function () { return Reflect.set(target, p, value, receiver); }, onSet);
        },
    });
}
exports.observeWeakMap = observeWeakMap;
var arrayMethodKeys = Reflect.ownKeys(Array.prototype);
var getObserver = function (p) {
    return arrayMethodKeys.indexOf(p) === -1
        ? observe
        : function (obj) { return obj; };
};
function observeArray(arr, onSet) {
    var createAtomIfNotExists = (0, atoms_1.getAtomCreator)(arr);
    return new Proxy(arr, {
        // TODO: fix `symbol` type
        get: function (target, p, receiver) {
            var atom = createAtomIfNotExists(p);
            return (0, calculations_1.calculate)(atom, function () { return Reflect.get(target, p, receiver); }, getObserver(p));
        },
        // TODO: fix `symbol` type
        set: function (target, p, value, receiver) {
            var atom = createAtomIfNotExists(p, value);
            return (0, mutations_1.mutate)(atom, function () { return Reflect.set(target, p, value, receiver); }, onSet);
        },
        // TODO: fix `symbol` type
        defineProperty: function (target, p, attributes) {
            var value = attributes.value;
            createAtomIfNotExists(p, value);
            return Reflect.defineProperty(target, p, attributes);
        },
    });
}
exports.observeArray = observeArray;
function observe(obj, onSet) {
    switch (typeof obj) {
        case 'object':
            if (obj === null)
                return obj;
            if (obj instanceof Observable_1.Observable)
                return obj;
            if (obj instanceof Promise)
                return obj;
            if (obj instanceof WeakMap)
                return obj;
            if (Array.isArray(obj))
                return observeArray(obj, onSet);
            return observeObj(obj, onSet);
        case 'function':
            return observeFunction(obj, onSet);
        default:
            return obj;
    }
}
exports.observe = observe;
//# sourceMappingURL=observe.js.map